#include "string.h"
#include "esp_types.h"
#include "esp_attr.h"
#include "esp_intr.h"
#include "esp_err.h"
#include "esp_log.h"
#include "freertos/FreeRTOS.h"
#include "freertos/semphr.h"
#include "freertos/queue.h"
#include "freertos/xtensa_api.h"
#include "soc/uhci_reg.h"
#include "soc/uhci_struct.h"
#include "driver/periph_ctrl.h"
#include "freertos/ringbuf.h"
#include "driver/udma.h"


#define TIMEOUT_DEFAULT_SHIFT       (0x0)       

#define UHCI_NUM_ERR_STR            "uhci_num error!!! valid param: { UHCI_NUM_0, UHCI_NUM_0 }"
#define UHCI_LINK_TYPE_ERR_STR      "uhci_link_type error!!! valid param: { UHCI_LINK_IN, UHCI_LINK_OUT }"


#define UHCI_ENTER_CRITICAL(mux)    portENTER_CRITICAL(mux)
#define UHCI_EXIT_CRITICAL(mux)     portEXIT_CRITICAL(mux)

#define UHCI_CHECK(a, str, ret_val) \
    if (!(a)) { \
        ESP_LOGE(UHCI_TAG,"%s(%d): %s", __FUNCTION__, __LINE__, str); \
        return (ret_val); \
    }


static const char* UHCI_TAG = "uhci module";
static DRAM_ATTR uhci_dev_t * const UHCI[UHCI_NUM_MAX] = {&UHCI0, &UHCI1};
static portMUX_TYPE uhci_spinlock[UHCI_NUM_MAX] = {portMUX_INITIALIZER_UNLOCKED, portMUX_INITIALIZER_UNLOCKED};

extern QueueHandle_t uhci_queue;


static struct uhci_dev_t {
    uhci_num_t uhci_num;                 //UHCI num
    int queue_size;                      //queue size 
    QueueHandle_t xQueueUhci;
    SemaphoreHandle_t tx_sem;
    uhci_isr_handle_t  uhci_isr_handle;  //isr handle     
    uint32_t rx_data_len;                //data len
    uhci_dma_link_t rx_link;             //out description link
    uhci_dma_link_t tx_link;             //in  description link
} *uhci_dev[UHCI_NUM_MAX] = {NULL};



esp_err_t uhci_clear_intr_status(uhci_num_t uhci_num, uint32_t clr_mask)
{
    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);

    UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
    SET_PERI_REG_MASK(UHCI_INT_CLR_REG(uhci_num), clr_mask);
    UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    return ESP_OK;   
}


esp_err_t uhci_intr_enable(uhci_num_t uhci_num, uint32_t ena_mask)
{
    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);

    UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
    SET_PERI_REG_MASK(UHCI_INT_ENA_REG(uhci_num), ena_mask);
    UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    return ESP_OK; 
}


esp_err_t uhci_intr_disable(uhci_num_t uhci_num, uint32_t ena_mask)
{
    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);

    UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
    CLEAR_PERI_REG_MASK(UHCI_INT_ENA_REG(uhci_num), ena_mask);
    UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    return ESP_OK;   
}


esp_err_t uhci_set_link_addr(uhci_num_t uhci_num, uhci_link_type_t link_type, uint32_t link_addr)
{
	UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);
    UHCI_CHECK(link_type < UHCI_LINK_MAX, UHCI_LINK_TYPE_ERR_STR, ESP_ERR_INVALID_ARG);

    if(link_type == UHCI_LINK_OUT) {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        UHCI[uhci_num]->dma_out_link.addr = (link_addr & UHCI_OUTLINK_ADDR_M);
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    } else {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        UHCI[uhci_num]->dma_in_link.addr = (link_addr & UHCI_OUTLINK_ADDR_M);
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    }
    return ESP_OK;	
}



esp_err_t uhci_link_start(uhci_num_t uhci_num, uhci_link_type_t link_type)
{
    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);
    UHCI_CHECK(link_type < UHCI_LINK_MAX, UHCI_LINK_TYPE_ERR_STR, ESP_ERR_INVALID_ARG);

    if(link_type == UHCI_LINK_OUT) {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        UHCI[uhci_num]->dma_out_link.start = 1;
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    } else {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        UHCI[uhci_num]->dma_in_link.start = 1;
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    }
    return ESP_OK;   
}


esp_err_t uhci_link_stop(uhci_num_t uhci_num, uhci_link_type_t link_type)
{
    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);
    UHCI_CHECK(link_type < UHCI_LINK_MAX, UHCI_LINK_TYPE_ERR_STR, ESP_ERR_INVALID_ARG);

    if(link_type == UHCI_LINK_OUT) {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        UHCI[uhci_num]->dma_out_link.stop = 1;
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    } else {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        UHCI[uhci_num]->dma_in_link.stop = 1;
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    }
    return ESP_OK;   
}


esp_err_t uhci_link_restart(uhci_num_t uhci_num, uhci_link_type_t link_type)
{
    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);
    UHCI_CHECK(link_type < UHCI_LINK_MAX, UHCI_LINK_TYPE_ERR_STR, ESP_ERR_INVALID_ARG);

    if(link_type == UHCI_LINK_OUT) {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        UHCI[uhci_num]->dma_out_link.restart = 1;
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    } else {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        UHCI[uhci_num]->dma_in_link.restart = 1;
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    }
    return ESP_OK;   
}


esp_err_t uhci_link_reset(uhci_num_t uhci_num, uhci_link_type_t link_type)
{
    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);
    UHCI_CHECK(link_type < UHCI_LINK_MAX, UHCI_LINK_TYPE_ERR_STR, ESP_ERR_INVALID_ARG);

    if(link_type == UHCI_LINK_OUT) {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        UHCI[uhci_num]->conf0.out_rst = 1;
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    } else {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        UHCI[uhci_num]->conf0.in_rst = 1;
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    }
    return ESP_OK;   
}


esp_err_t uhci_get_link_park(uhci_num_t uhci_num, uhci_link_type_t link_type, int *status)
{
    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);
    UHCI_CHECK(link_type < UHCI_LINK_MAX, UHCI_LINK_TYPE_ERR_STR, ESP_ERR_INVALID_ARG);

    if(link_type == UHCI_LINK_OUT) {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        *status = UHCI[uhci_num]->dma_out_link.park;
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    } else {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        *status = UHCI[uhci_num]->dma_in_link.park;
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    }
    return ESP_OK;   
}


esp_err_t unci_isr_register(uhci_num_t uhci_num, void (*fun)(void*), void *arg, int intr_alloc_flags, uhci_isr_handle_t *handle)
{
    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);

    esp_err_t ret;
    
    if(uhci_num == UHCI_NUM_1) {
        ret = esp_intr_alloc(ETS_UHCI1_INTR_SOURCE, intr_alloc_flags, fun, arg, handle);
    } else {
        ret = esp_intr_alloc(ETS_UHCI0_INTR_SOURCE, intr_alloc_flags, fun, arg, handle);
    }
    return ret;
}


esp_err_t uhci_attach_uart(uhci_num_t uhci_num, uint32_t uart_num)
{
    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);

    if(uart_num == 1) {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        UHCI[uhci_num]->conf0.uart1_ce = 1;
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);    
    }else if(uart_num == 2) {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        UHCI[uhci_num]->conf0.uart2_ce = 1;
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    } else {
        UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
        UHCI[uhci_num]->conf0.uart0_ce = 1;
        UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    }
    return ESP_OK;
}



esp_err_t uhci_config(uhci_config_t *config)
{
    uhci_num_t uhci_num = config->uhci_num;
    unsigned int uart_num = config->uart_num;
    uint32_t intr_mask = config->intr_mask;

    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);

    periph_module_enable(PERIPH_UHCI0_MODULE);
    UHCI[uhci_num]->conf0.val = 0x0;
    UHCI[uhci_num]->conf1.val = 0x0;
    UHCI[uhci_num]->escape_conf.val = 0x0;
    UHCI[uhci_num]->hung_conf.val = 0x0;
    UHCI[uhci_num]->conf0.uart_idle_eof_en = 1;
    //UHCI[uhci_num]->conf0.clk_en = 1;
    uhci_attach_uart(uhci_num, uart_num);
    return ESP_OK;
}


static void  uhci_intr_handle(void *parm)
{
    struct uhci_dev_t *uhci = (struct uhci_dev_t  *)parm;
    uhci_event_t uhci_event;
    
    UHCI_ENTER_CRITICAL(&uhci_spinlock[UHCI_NUM_0]);

    uint32_t status = UHCI[UHCI_NUM_0]->int_st.val;
    UHCI_EXIT_CRITICAL(&uhci_spinlock[UHCI_NUM_0]);
    uhci_event.intr_status = status;
    uhci_clear_intr_status(UHCI_NUM_0, status);

    if(status & UHCI_RX_START) {
        uhci_event.event_type = RX_START;          
    } else if(status & UHCI_TX_START) {
        uhci_event.event_type = TX_START;
    } else if(status & UHCI_RX_HUNG) {
        uhci_event.event_type = RX_HUNG;
    } else if(status & UHCI_TX_HUNG) {
        uhci_event.event_type = TX_HUNG;
    } else if(status & UHCI_OUT_DONE) {
        uhci_event.event_type = OUT_DONE;
    } else if(status & UHCI_OUT_EOF) {
        uhci_event.event_type = OUT_DONE;
        xSemaphoreGive(uhci->tx_sem);
    } else if(status & UHCI_IN_DONE) {
        uhci_event.event_type = IN_DONE;
        
    } else {
        uhci_event.event_type = UNDEFINED;    
    }

    if(uhci->xQueueUhci != NULL) 
        xQueueSendFromISR(uhci->xQueueUhci, (void * )&uhci_event,  (portBASE_TYPE) 0);
}


esp_err_t uhci_set_thr_val(uhci_num_t uhci_num, uint8_t thr_val)
{
    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);
    
    UHCI_ENTER_CRITICAL(&uhci_spinlock[uhci_num]);
    UHCI[uhci_num]->conf1.dma_in_fifo_full_thrs = thr_val;
    UHCI_EXIT_CRITICAL(&uhci_spinlock[uhci_num]);
    return ESP_OK;
}


esp_err_t uhci_out_push(uhci_num_t uhci_num, uint8_t data)
{
    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);

    UHCI[uhci_num]->dma_out_push.fifo_push = 0;
    UHCI[uhci_num]->dma_out_push.fifo_wdata = data;
    UHCI[uhci_num]->dma_out_push.fifo_push = 1;

    return ESP_OK;   
}


esp_err_t uhci_set_time_out(uhci_num_t uhci_num, uint8_t tx_timeout, uint8_t rx_timeout)
{
    
    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);
    UHCI[uhci_num]->hung_conf.txfifo_timeout = tx_timeout;
    UHCI[uhci_num]->hung_conf.rxfifo_timeout = rx_timeout;
    UHCI[uhci_num]->hung_conf.rxfifo_timeout_shift = 0;
    UHCI[uhci_num]->hung_conf.rxfifo_timeout_shift = 0;
    UHCI[uhci_num]->hung_conf.txfifo_timeout_ena = 1;
    UHCI[uhci_num]->hung_conf.rxfifo_timeout_ena = 1;
    UHCI[uhci_num]->conf0.uart_idle_eof_en = 1;
    return ESP_OK;    
}


esp_err_t uhci_driver_install(uhci_num_t uhci_num, QueueHandle_t *uhci_queue, int rx_buf_size, int queue_size, int intr_alloc_flags)
{
    UHCI_CHECK(uhci_num < UHCI_NUM_MAX, UHCI_NUM_ERR_STR, ESP_ERR_INVALID_ARG);

    if(uhci_dev[uhci_num] != NULL)
        return ESP_FAIL;
    uhci_dev[uhci_num] = (struct uhci_dev_t *)malloc(sizeof(struct uhci_dev_t));

    if(uhci_dev[uhci_num] == NULL) {
        return ESP_FAIL;
    }

    uhci_dev[uhci_num]->uhci_num = uhci_num;
    uhci_dev[uhci_num]->queue_size = queue_size;

    vSemaphoreCreateBinary(uhci_dev[uhci_num]->tx_sem);

    if(uhci_queue != NULL) {
        uhci_dev[uhci_num]->xQueueUhci = xQueueCreate(queue_size, sizeof(uhci_event_t));
        *uhci_queue = uhci_dev[uhci_num]->xQueueUhci;
    } else {
        uhci_dev[uhci_num]->xQueueUhci = NULL; 
    }

    uhci_dev[uhci_num]->rx_link.des.size = 0;
    uhci_dev[uhci_num]->rx_link.des.length = 0;
    uhci_dev[uhci_num]->rx_link.des.eof = 1;
    uhci_dev[uhci_num]->rx_link.des.owner = 1;
    uhci_dev[uhci_num]->rx_link.buf = NULL;
    uhci_dev[uhci_num]->rx_link.pnext = NULL;

    uhci_dev[uhci_num]->tx_link.des.size = 0;
    uhci_dev[uhci_num]->tx_link.des.length = 0;
    uhci_dev[uhci_num]->tx_link.des.eof = 1;
    uhci_dev[uhci_num]->tx_link.des.owner = 1;
    uhci_dev[uhci_num]->tx_link.buf = NULL;
    uhci_dev[uhci_num]->tx_link.pnext = NULL;

    uhci_set_link_addr(uhci_num, UHCI_LINK_IN, (uint32_t)&(uhci_dev[uhci_num]->rx_link));
    uhci_set_link_addr(uhci_num, UHCI_LINK_OUT, (uint32_t)&(uhci_dev[uhci_num]->tx_link));
    unci_isr_register(uhci_num, uhci_intr_handle, (void *)uhci_dev[uhci_num], intr_alloc_flags, &(uhci_dev[uhci_num]->uhci_isr_handle));
    return ESP_OK;
}


esp_err_t uhci_write_data(uhci_num_t uhci_num, const uint8_t *data_buf, uint32_t data_size)
{
    if(uhci_dev[uhci_num] == NULL)
        return ESP_FAIL;
    uhci_dev[uhci_num]->tx_link.des.length = data_size;
    uhci_dev[uhci_num]->tx_link.buf = data_buf;
    xSemaphoreTake(uhci_dev[uhci_num]->tx_sem, (TickType_t) 0);
    uhci_link_start(uhci_num, UHCI_LINK_OUT);
    return ESP_OK;
}


esp_err_t uhci_read_data(uhci_num_t uhci_num, uint8_t *data_buf, uint32_t buf_size)
{
    if(uhci_dev[uhci_num] == NULL)
        return ESP_FAIL;
    uhci_dev[uhci_num]->rx_link.des.size = buf_size;
    uhci_dev[uhci_num]->rx_link.buf = data_buf;
    uhci_link_start(uhci_num, UHCI_LINK_IN);
    return ESP_OK;
}
