#ifndef __DRIVER_UHCI_H__
#define __DRIVER_UHCI_H__

#ifdef __cplusplus
extern "C" {
#endif

#include "freertos/semphr.h"
#include "freertos/queue.h"
#include "esp_intr_alloc.h"
#include "esp_intr.h"
#include "soc/uhci_struct.h"
#include "soc/uhci_reg.h"
#include "esp_err.h"


typedef intr_handle_t uhci_isr_handle_t;

typedef enum {
    UHCI_NUM_0  = 0x0,
    UHCI_NUM_1  = 0x1,
    UHCI_NUM_MAX,
} uhci_num_t;


typedef enum{
    UHCI_LINK_IN  = 0x0,
    UHCI_LINK_OUT = 0x1,
    UHCI_LINK_MAX,
} uhci_link_type_t;


typedef enum {
    UHCI_INTR_DIS          = (0x0),                // disable all uhci related interrupts
    UHCI_RX_START          = (0X1 << 0),           // rx_start interrupt mask
    UHCI_TX_START          = (0X1 << 1),           // tx_start interrupt mask
    UHCI_RX_HUNG           = (0X1 << 2),           // rx_hung interrupt mask
    UHCI_TX_HUNG           = (0X1 << 3),           // tx_hung interrupt mask
    UHCI_IN_DONE           = (0X1 << 4),           // in discreaption link complited
    UHCI_IN_SUC_EOF        = (0X1 << 5),           
    UHCI_IN_ERR_EOF        = (0X1 << 6),
    UHCI_OUT_DONE          = (0X1 << 7),
    UHCI_OUT_EOF           = (0X1 << 8),
    UHCI_IN_DSCR_ERR       = (0X1 << 9),
    UHCI_OUT_DSCR_ERR      = (0X1 << 10),
    UHCI_IN_DSCR_EMPTY     = (0X1 << 11),
    UHCI_OUTLINK_EOF_ERR   = (0X1 << 12),
    UHCI_OUT_TOTAL_EOF     = (0X1 << 13),  
    UHCI_SEND_S_REG_Q      = (0X1 << 14),
    UHCI_SEND_A_REG_Q      = (0X1 << 15),
    UHCI_DMA_IN_FIFO_FULL  = (0X1 << 16),
    UHCI_INTR_MAX,
} uhci_intr_type_t;


typedef enum{
    UNDEFINED = 0,
    RX_START,
    TX_START,
    RX_HUNG,
    TX_HUNG,
    OUT_DONE,
    OUT_EOF,
    IN_DONE,
} uhci_event_type_t;


typedef struct {
    uhci_event_type_t event_type;
    uint32_t data_len;
    uint32_t send_sta;

} uhci_event_t;


typedef struct {
    uhci_num_t  uhci_num;
    uint32_t uart_num;
    uint32_t rx_time_out;
    uint32_t tx_time_out;
    uint32_t intr_mask;
} uhci_config_t;


typedef struct uhci_dma_link {
    struct {
        uint32_t size :    12;      //the size of buf, must be able to be divisible by 4 
        uint32_t length:   12;      //in link, 
        uint32_t reversed: 6;       //reversed
        uint32_t eof:      1;       //if this dma link is the last one, you shoule set this bit 1.      
        uint32_t owner:    1;       //the owner of buf, bit 1 : DMA, bit 0 : CPU.
    } des;
    uint8_t *buf;                   //the pointer of buf
    struct uhci_dma_link *pnext;    //point to the next dma linke, if this link is the last one, set it NULL.
} uhci_dma_link_t;


extern  esp_err_t uhci_clear_intr_status(uhci_num_t uhci_num, uint32_t clr_mask);

extern  esp_err_t uhci_intr_enable(uhci_num_t uhci_num, uint32_t ena_mask);

extern  esp_err_t uhci_intr_disable(uhci_num_t uhci_num, uint32_t ena_mask);

extern  esp_err_t uhci_enable_rx_esc(uhci_num_t uhci_num, uint8_t esc);

extern  esp_err_t uhci_set_link_addr(uhci_num_t uhci_num, uhci_link_type_t link_type, uint32_t link_addr);

extern  esp_err_t uhci_get_link_park(uhci_num_t uhci_num, uhci_link_type_t link_type, int *status);

extern  esp_err_t uhci_link_start(uhci_num_t uhci_num, uhci_link_type_t link_type);

extern  esp_err_t uhci_link_stop(uhci_num_t uhci_num, uhci_link_type_t link_type);

extern  esp_err_t uhci_link_restart(uhci_num_t uhci_num, uhci_link_type_t link_type);

extern  esp_err_t uhci_isr_install(uhci_isr_handle_t handle, uhci_num_t uhci_num, int intr_alloc_flags, void *parm);

extern  esp_err_t uhci_config(uhci_config_t *config);

extern  esp_err_t uhci_attach_uart(uhci_num_t uhci_num, uint32_t uart_num);

extern  esp_err_t uhci_link_reset(uhci_num_t uhci_num, uhci_link_type_t link_type);

extern  esp_err_t uhci_in_pop(uhci_num_t uhci_num);

extern  esp_err_t uhci_set_time_out(uhci_num_t uhci_num, uint8_t tx_timeout, uint8_t rx_timeout);

extern  esp_err_t uhci_write_data(uhci_num_t uhci_num, const uint8_t *data_buf, uint32_t data_size);

extern  esp_err_t uhci_driver_install(uhci_num_t uhci_num, QueueHandle_t *uhci_queue, int queue_size, int intr_alloc_flags);

extern int uhci_read_data(uhci_num_t uhci_num, uint8_t *data_buf, uint32_t read_size);

extern int uhci_get_data_len(uhci_num_t uhci_num);

#ifdef __cplusplus__
}
#endif

#endif